#!/usr/bin/python

# Copyright 2022- H Fisher, Canberra
# Released under MIT open source license
# All bugs, suggestions, and other feedback to hugo.fisher@gmail.com

# Utility program to run a Python program to generate sequence of frames.

# How to use: rmframes.py program-name.py num-frames [ -out file-name-pattern ]

# Python program should have a render(**args) function

# Default output dir is /tmp/program-name/ or platform equivalent
# Default frame files are program-name#f.tiff
# -out arg can be just a directory or directory/filename pattern

# DESIGN NOTES

# No command line options for resolution, format, pixel aspect ratio etc
# You probably have standard settings for these. Make a copy and edit
# the code if you need to make changes for a particular project.

# No error recovery. If things go wrong, it just stops. Good enough for
# small projects, and if you're doing anything more complicated you
# presumably have your own Master Control Program to manage things.

# No post-processing. Same reason as no error recovery.

import sys, os, importlib, builtins, math, tempfile, string, fnmatch, time
from os import path
import logging

import prman

# Our global settings
prog = None
nFrames = None
output = None

# Might add options for these one day
first   = 1
format  = "rgb"
FPS     = 30


def configFrame():
    """Resolution, pixel aspect ratio, etc"""
    Ri.DisplayFilter("PxrBackgroundDisplayFilter", "background", {
                        "color backgroundColor": [ 0, 0, 0 ] })
    Ri.Format(1280, 720, 1)
    Ri.Integrator("PxrDefault", "integrator")
    # TODO Final render much more likely to use full raytracing
    # Insert other stuff here...


def digits(n):
    """Return number of digits needed for frame numbers up to n"""
    # Need +1 because counting up to and including
    return int(math.ceil(math.log10(n + 1)))

def defaultFilePattern(prog):
    """Program hello.py defaults to frames hello#f.tiff"""
    name = path.splitext(path.basename(prog.__name__))[0]
    return name + "#" + str(digits(nFrames)) + "f.tiff"

def replaceFrameNum(fileName, newStr):
    """Replace #f or #Nf in fileName with newStr"""
    # Could probably do this in one line, I'm not expert so prefer simple steps
    i = fileName.index('#')
    prefix = fileName[0:i]
    suffix = fileName[i + 1:]
    while suffix[0] == 'f' or suffix[0].isdigit():
        suffix = suffix[1:]
    return prefix + newStr + suffix

def delMatchingFiles(d, framePattern):
    """Zot all files in d that match framePattern"""
    # Glob style wildcard for filenames
    wildcard = replaceFrameNum(framePattern, "*")
    # Now find all matching files
    targets = fnmatch.filter(os.listdir(d), wildcard)
    if len(targets) > 0:
        logging.info("Deleting frames from earlier run")
        for f in targets:
            os.remove(path.join(d, f))

def checkNFrames(arg):
    """Must be an int and not out of range"""
    n = int(arg)
    if n <= 0:
        raise ValueError("Number of frames must be positive")
    if n > 10000:
        raise ValueError(str(n) + " frames? Edit " + __file__ + " if you are serious")
    return n

def checkPython(srcName):
    """Import and return Python module object for file name"""
    modName = path.splitext(path.basename(srcName))[0]
    modDir = path.dirname(srcName)
    # Modifying the search path could maybe cause problems for other
    # modules later on, but I haven't encountered any yet
    if modDir:
        sys.path.append(modDir)
    mod = importlib.import_module(modName)
    return mod

def checkFrameOutput(target):
    """Try to ensure we can write frames"""
    # Sanity check on names
    if not '#' in path.basename(target) and nFrames > 1:
        raise RuntimeError("No #f in " + target + " for frame number")
    # Do not encourage Mac/MSWin bad habits in naming files
    for c in string.whitespace:
        if c in target:
            raise RuntimeError("Special characters in " + target + " ? Edit " + __file__ + " if you want to risk it")
    # Check directory
    d = path.dirname(target)
    if path.exists(d):
        if not path.isdir(d):
            raise IOError(d + " is not a directory")
        elif not os.access(d, os.W_OK):
            raise IOError("Cannot write files to " + d)
        else:
            # Remove any files from previous run
            delMatchingFiles(d, path.basename(target))
    else:
        # New directory. Assume intentional, not a mistake
        os.makedirs(d)


##  Main
##  I am trying to always have a function rather than inline code, because
##  this allows automatic documentation generators to import the program
##  without getting some kind of runtime error

def main():
    global prog, nFrames, output, first, format, FPS

    if len(sys.argv) < 3:
        raise RuntimeError("Usage: " + __file__ + " program-name.py num-frames [ options ]")

    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

    # Arguments
    idx = 1
    while idx < len(sys.argv):
        arg = sys.argv[idx]
        if not arg.startswith('-'):
            if prog is None:
                prog = checkPython(arg)
            elif nFrames is None:
                nFrames = checkNFrames(arg)
            else:
                raise RuntimeError("Unknown argument: " + arg)
        else:
            if arg == "-out":
                output = sys.argv[idx + 1]
                idx += 1
            else:
                raise RuntimeError("Unknown option: " + arg)
        idx += 1

    # Fill in any missing bits for dir/file output
    if output is None:
        # Default is SOMEWHERE/name/name#f.tiff, allowing for platform differences and big frame counts
        name = path.splitext(path.basename(prog.__name__))[0]
        # Linux and MacOS have well defined /tmp dir. On MS Windows, default
        # temp dir is \Users\NAME\AppData\Local\Temp which is invisible to
        # regular GUI apps, so create our own Pictures\RMan\
        if sys.platform.lower().startswith("win32"):
            tempDir = path.join(path.expanduser("~"), "Pictures", "RenderMan")
            try:
                os.mkdir(tempDir)
            except OSError:
                pass    # Already exists
        else:
            tempDir = tempfile.gettempdir()
        output = path.join(tempDir, name, defaultFilePattern(prog))
    elif path.isdir(output):
        # Dir only, add default name pattern
        output = path.join(output, defaultFilePattern(prog))
    # And make sure everything is OK
    logging.info("Frames to {}".format(output))
    checkFrameOutput(output)

    # Finally, ready to roll
    builtins.__dict__["Ri"] = prman.Ri()
    Ri.Begin(Ri.RENDER)

    for frame in range(first, first + nFrames):
        logging.info("Frame {:d}".format(frame))
        Ri.FrameBegin(frame)
        # 8 bit channel output so can use ffmpeg and standard paint apps
        Ri.Display(output, "file", format, { "string format": "uint8"})
        configFrame()
        prog.render(frame=frame,
                filename=output,
                duration=nFrames,
                fps=FPS)
        Ri.FrameEnd()
        # Synch so any RenderMan errors will match printed frame number
        prman.RicFlush(Ri.EMPTY, True, Ri.FINISHRENDERING)

    Ri.End()

    # Print dir name again as reminder
    logging.info("Done. Frames written to {}".format(path.dirname(output)))


if __name__ == "__main__":
    main()
