#!/usr/bin/python

# Copyright 2022- H Fisher, Canberra
# Released under MIT open source license
# All bugs, suggestions, and other feedback to hugo.fisher@gmail.com

# Utility program to re-run a RenderMan Python program every time
# the program (or dependencies) changes.

# How to use: rmwatch.py python-program [ file pattern ] ...

# Python program should have a render(**args) function

# Simple case is to run rmwatch in a terminal while editing your python program
#       % rmwatch.py my-renderman.py
# Every time you save changes, rmwatch will re-execute my-renderman.py

# Advanced case is something like
#       % rmwatch.py my-renderman.py *.oso *.tex
# which will also detect when you recompile shaders or alter textures

# DESIGN NOTES

# It isn't very smart. Won't auto detect changes to imported Python modules
# or RIB Archives unless you explicitly add *.py or *.rib patterns. This is
# meant for playing around with simple programs, not controlling your entire
# production pipeline. You can always hit Save for the main program to force
# a new render.

# Need to run within the right working directory for the Python program.
# I experimented with checking if the Python program was in a different
# dir and cd'ing into it, but the added complexity didn't seem worth it.

# I thought about adding a couple of keystroke commands, to force a render
# and to quit. That needs asynchronous IO or another thread, and preferably
# single char rather than line input, and again it didn't seem worth it.

import sys, os, time, traceback, importlib, builtins
from os import path
import logging

import prman


def configFrame():
    """Resolution, integrator for lighting, etc"""
    Ri.Format(1280, 720, 1)
    ## Very simple light source at camera
    Ri.Integrator("PxrDefault", "integrator")
    # TODO command line args to select integrator?
    ## Visualizer is good for working on geometry
    #Ri.Integrator("PxrVisualizer", "integrator")
    ## Full raytracing. Needs at least one light source in scene
    # Got these parameters from Pixar example scene
    # Ri.Hider("raytrace", {
    #         "int incremental": 1,
    #         "int minsamples": 16,
    #         "int maxsamples": 256,
    #     })
    # # Cut down bounces from default
    # Ri.Integrator("PxrPathTracer", "integrator", {
    #         "int maxIndirectBounces": 4,
    #     })
    ## Insert other stuff here...


Modified = {}    # Will be last modification time for files

def checkForChanges(patterns):
    """Return true if any files matched by patterns have been modified"""
    global Modified
    # Update all time stamps, not just first. Otherwise edits to more
    # than one file at a time would trigger successive re-renders
    changed = False
    for s in patterns:
        t = path.getmtime(s)
        if s not in Modified:
            Modified[s] = t
            changed = True
        elif Modified[s] != t:
            logging.debug("{} changed".format(s))
            Modified[s] = t
            changed = True
    return changed

def loadPython(srcName):
    """Import and return Python module object for file name"""
    modName = path.splitext(path.basename(srcName))[0]
    modDir = path.dirname(srcName)
    if modDir and modDir not in sys.path:
        sys.path.append(modDir)
        logging.info("Added {} to sys.path".format(modDir))
    try:
        logging.info("Load {}".format(srcName))
        mod = importlib.import_module(modName)
    except Exception as e:
        logging.warning("Main program error {}".format(e))
        mod = None
    return mod

def reloadProgram(targets, mainName, prog):
    """Reload watched modules and program"""
    try:
        # Reload imports first
        for f in targets:
            if f != mainName and f.endswith(".py"):
                mod = sys.modules[path.splitext(path.basename(f))[0]]
                logging.info("Reload dependency {}".format(f))
                importlib.reload(mod)
        prog = importlib.reload(prog)
    except Exception as e:
        logging.warning("Reloading error {}".format(e))
        prog = None
    return prog

def resetPreview():
    """Clean up RenderMan state in case previous run had an error"""
    try:
        Ri.ErrorHandler(Ri.ErrorIgnore)
        Ri.FrameBegin(1)
        try:
            Ri.WorldBegin()
        finally:
            Ri.WorldEnd()
    finally:
        Ri.FrameEnd()
        Ri.ErrorHandler(Ri.ErrorPrint)

def runProgram(prog):
    """Run the program and display result of render"""
    # Don't stop if program is wrong
    try:
        # Window title is origin Python plus timestamp
        timeStamp = time.strftime("%H:%M:%S", time.localtime())
        fName = path.basename(prog.__file__) + "-" + timeStamp
        logging.info("{} Render {}...".format(timeStamp, prog.__name__))
        Ri.Display(fName, "framebuffer", "rgb")
        Ri.DisplayFilter("PxrBackgroundDisplayFilter", "background", {
                        "color backgroundColor": [ 0, 0, 0 ] })
        Ri.FrameBegin(1)
        configFrame()
        prog.render(frame=1, filename=fName, FPS=30, shutter=0.025)
    except Exception as e:
        # Error running code, not importing, so print full stack trace
        traceback.print_tb(sys.exc_info()[2])
        logging.warning(e)
        prog = None
    # Synch with renderer
    Ri.FrameEnd()
    prman.RicFlush(Ri.EMPTY, True, Ri.FINISHRENDERING)
    return prog


##  Main
##  I am trying to always have a function rather than inline code, because
##  this allows automatic documentation generators to import the program
##  without getting some kind of runtime error

def main():
    global Modified

    if len(sys.argv) == 1:
        raise RuntimeError("Usage: " + __file__ + " python-program [ file pattern ] ...")

    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

    # Program we're going to run
    targetFile = sys.argv[1]
    if not (path.isfile(targetFile)):
        raise RuntimeError("Cannot find " + targetFile)
    # If prog is not Python, error will be reported on first render
    prog = None

    logging.info("Program: {}".format(targetFile))
    logging.info("Re-render on changes to: {}".format(' '.join(sys.argv[1:])))

    # Now just run forever or until we get Ctrl-C'ed

    # I'm going to the special hell for this... TOOD find better way
    builtins.__dict__["Ri"] = prman.Ri()

    Ri.Begin(Ri.RENDER)
    while True:
        if checkForChanges(sys.argv[1:]):
            resetPreview()
            if prog is None:
                prog = loadPython(targetFile)
            else:
                # Reloading multiple modules in Python is generally not recommended
                # and can go wrong for many, many reasons. Easiest solution if not
                # working is just to Ctrl-C rmwatch.py and restart
                prog = reloadProgram(Modified.keys(), targetFile, prog)
            # Old version had horrible platform specific code to shut down and
            # restart the RMan preview. Now the previewer stays open ready to
            # display new frames. Thank you Pixar!
            if prog is not None:
                runProgram(prog)
        # This is busy waiting, because at time of writing Python doesn't
        # have a standard module that waits for file system change events.
        # We're between renders, so doesn't hurt to use some CPU cycles.
        time.sleep(1)
    Ri.End()

if __name__ == "__main__":
    main()

